/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * MooDS: a communication protocol used in GASP.
 * GASP Demonstrator: a gasp demonstration game.
 *
 * Copyright (C) 2004 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Franck Caron, Romain Pellerin
 */

//#include parameters.h
////////////////////////
// Game params        //
////////////////////////

////#define PING
////#define VIRTUAL_WORLD_MAP
//#endinclude

package org.ugasp.games.demonstrator;


//import java.util.Enumeration;

import java.awt.Point;
import java.util.Enumeration;
import java.util.Hashtable;

//import org.apache.log4j.Category;
import org.ugasp.core.ApplicationInstance;
import org.ugasp.eventmodel.event.*;
import org.ugasp.eventmodel.listener.*;
import org.ugasp.eventmodel.event.impl.*;

//#ifdef PING 
//# import org.ugasp.demonstrator.PingPlayer;
//#endif PING

//#ifdef VIRTUAL_WORLD_MAP
//# import org.ugasp.demonstrator.UpdatePlayer;
//#endif VIRTUAL_WORLD_MAP

import org.ugasp.protocol.moods.CustomTypes;
import org.ugasp.service.internal.gameserver.location.Projection;

/**
 * @author PELLERIN Romain (CNAM Paris, romain.pellerin@laposte.net)
 * @author DELPIANO Fabien (Pastagames, fabien@pastaga.net)
 * @author KOCIK Fabrice - (Pastagames, fabrice@pastis.org)
 * @author CARON Franck - (CNAM Paris)
 * --
 * JIMM Project 2005
 * (Bouygues Telecom, CNAM Paris, Pastagames)
 */
public class Server implements org.ugasp.gameserver.Server, OnDataEvent, OnEndEvent, OnJoinEvent, OnQuitEvent, OnStartEvent  {
	/*****************************************************/
	/** Mandatory code for all uGASP based game servers **/
	/*****************************************************/
	
	// iPojo injection fields
	private Projection projection;
	
	// Other fields
    protected ApplicationInstance appIns;
	private org.ugasp.protocol.moods.CustomTypes customTypes;
	private org.ugasp.protocol.moods.string.CustomTypes customTypes4StringEncoding;
	
    /**
     * Link the application instance to the game server logic.
     * 
     * @param applicationInstance
     */
    public void setOwnerAI(ApplicationInstance applicationInstance){
        appIns = applicationInstance;
    }
    
    
    /**
     * Send to the linked applicationInstance a DataEvent for a specific actorSession.
     * 
     * @param actorSessionID
     * @param dataEvent the DataEvent to send
     */
    public void sendDataTo(int actorSessionID, DataEvent dataEvent){
        appIns.sendDataToActorSession(actorSessionID,dataEvent);
    }
    
    
    /**
     * Retrieve the actor ID linked to the actor session ID.
     * (persistency) Provide the storage of player game informations. 
     * 
     * @param actorSessionID
     * @param dataEvent the DataEvent to send
     */
    public int retrieveActorID(int actorSessionID){
        return appIns.getActorSession(actorSessionID).getActorID();
    }
    
    
    /**
     * Retrieve the application ID.
     * (persistency) Provide the storage of player game informations. 
     */
    public int retrieveApplicationID(){
        return appIns.retrieveApplicationID();
    }
    
    @Override
	public org.ugasp.protocol.moods.string.CustomTypes getCustomTypes4StringEncoding() {
		// TODO Auto-generated method stub
		return customTypes4StringEncoding;
	}


	@Override
	public CustomTypes getCustomTypes() {
		// TODO Auto-generated method stub
		return customTypes;
	}
    
    /*****************************************************/
	/** Game server logic code						    **/
	/*****************************************************/
	
	private int maxActor = 0;
	private boolean gameIsStarted = false;
	
	/**
	 * This table allow us to know if a player is in game or not.
	 * players[aSID]=1 if the player is connected, else -1.
	 */

	private Hashtable playersTable = new Hashtable();
	
	/**
	 * This table allow us to store pings received from players to transmit to other players
	 */
	

	//#ifdef PING
	//# //private Hashtable htPlayerPing = new Hashtable();
	//#endif PING
	
	public Server() {
		//super();
		//thread = new Thread(this);
		customTypes = new CustomTypesImpl();
		
		// this server does not currently communicate with a string based client
		customTypes4StringEncoding = null;
		
		//init server
		init();

		//cat = Category.getInstance("GASPLogging");
		System.out.println("uGASP Demonstrator Server instanciated");
	}

	/**
	 * Initialization of the fields of the server.
	 */
	private void init() {}

	public void start() {
		System.out.println("Server START running!");
		//thread.start();
	}

	public void end() {
		System.out.println("Server STOP!");
		//thread.stop();
	}

	/*********************************************************************************/
	/** Listenners : onStartEvent, onEndEvent, onJoinEvent, onQuitEvent, onDataEvent**/
	/*********************************************************************************/

	public void onStartEvent(StartEvent se) {
		System.out.println("STARTEVENT received by server: aSID>"
				+ se.getActorSessionID());
		start();
		gameIsStarted = true;
		//players = new Player[maxActor + 1];
		//pings = new Vector[maxActor + 1];
		
		/*for(int i = 1; i <= maxActor; i++){
			pings[i] = new Vector();
		}*/
	}

	public void onEndEvent(EndEvent ee) {
		System.out.println("ENDEVENT received by server: aSID>"
				+ ee.getActorSessionID());
		end();
	}

	public void onJoinEvent(JoinEvent je) {
		System.out.println("JOINEVENT received by server: aSID>"
				+ je.getActorSessionID() + " username>" + je.getUsername());

		maxActor++;
		Player p = new Player(je.getActorSessionID());
		playersTable.put(""+je.getActorSessionID(),p);
	}
	
	public void onQuitEvent(QuitEvent qe) {
		System.out.println("QUITEVENT received by server: aSID>"
				+ qe.getActorSessionID());

		// for waiting room only
		if (!gameIsStarted)
			maxActor--;
		
		int aSID = qe.getActorSessionID();
		playersTable.remove(""+aSID);
		
		//#ifdef PING
		//# if (gameIsStarted) {
			//# htPlayerPing.remove(""+aSID);
			//# //players[aSID] = null;
		//# }
		//#endif PING

	}

	public void onDataEvent(DataEvent de) {
		System.out.println("DATAEVENT received by server: aSID>"
				+ de.getActorSessionID() + " Hashtable size:"
				+ de.getData().size());

		Hashtable h1 = de.getData();

		int nbObj = h1.size();

		for (int i = 0; i < nbObj; i++) {
			 Object o = (Object) h1.get(i + "");
			 treatMessage(o, de.getActorSessionID());
		 }
	}
	
	/**
	 * Treatment of Messages object received.
	 * 
	 */
	private synchronized void treatMessage(Object o, int senderASID) {
		 Player p = null;
		 if (o instanceof NewPlayer) {
			 NewPlayer np = (NewPlayer) o; 
			 System.out.println("*****************************************");
			 System.out.println("BEGIN of the NEW PLAYER TREATMENT");
			 System.out.println("Object New Player received:");
			 System.out.println("ID: " + np.getId());
			 System.out.println("X: " + np.getX());
			 System.out.println("Y: " + np.getY());
			 System.out.println("Username :"+ np.getUsername());
			 
			 p = (Player) playersTable.get(""+np.getId());
			 p.setX(np.getX());
			 p.setY(np.getY());
			 p.setUsername(np.getUsername());
			 
			 //System.out.println("Player informations received: " + p.getId());
			 //sendInGamePlayersInfos(np.getId());
			 notifyToAllOtherPlayers(np, np.getId());
			 
			 System.out.println("END of the NEW PLAYER TREATMENT");
			 System.out.println("*****************************************");
		 }
		 //#ifdef VIRTUAL_WORLD_MAP
		 //# else if (o instanceof UpdatePlayer){
			 //# UpdatePlayer updp = (UpdatePlayer) o; 
			 //# System.out.println("*****************************************");
			 //# System.out.println("BEGIN of the Update PLAYER TREATMENT");
			 //# System.out.println("Object Update Player received:");
			 //# System.out.println("ID: " + updp.getId());
			 //# System.out.println("X: " + updp.getX());
			 //# System.out.println("Y: " + updp.getY());
			 //# System.out.println("SpeedX: " + updp.getSpeedX());
			 //# System.out.println("SpeedY: " + updp.getSpeedY());
			//# 
			 //# p = (Player) playersTable.get(""+updp.getId());
			 //# //p = players[updp.getId()];
					 //# 
			//# if (p!=null){
			 //# 
				 //# // Speed and Position Update of the player updp
				 //# p.setX(updp.getX());
				 //# p.setY(updp.getY());
				 //# p.setSpeedX(updp.getSpeedX());
				 //# p.setSpeedY(updp.getSpeedY());			 
				 //# notifyToAllOtherPlayers(updp, updp.getId());
			 //# }
			 //# 
			 //# System.out.println("END of the Update PLAYER TREATMENT");
			 //# System.out.println("*****************************************");
		 //# }
		 //#endif VIRTUAL_WORLD_MAP
		 //#ifdef PING
		 //# else if (o instanceof PingPlayer){
			 //# PingPlayer pp = (PingPlayer) o;
			 //# System.out.println("*****************************************");
			 //# System.out.println("BEGIN of the Ping PLAYER TREATMENT");
			 //# System.out.println("Object Ping Player received:");
			 //# System.out.println("ID Ping: " + pp.getId());
			 //# System.out.println("Counter " + pp.getPingNum());
			 //# 
			 //# // timestamping
			 //# pp.setTimestamp(System.currentTimeMillis());		 
			 //# 
			 //# notifyToAllOtherPlayers (pp, pp.getId());
			 //# // add to each other players
			 //# addPingMessageToEachOtherPlayers(pp);
			 //# 
			 //# // send to player each other players pings
			 //# sendEachOtherPlayersPings(pp.getId(), System.currentTimeMillis());
			 //# System.out.println("END of the ping PLAYER TREATMENT");
			 //# System.out.println("*****************************************"); 
		 //# }
		//#endif PING
		if (o instanceof ChatMessageBroadcast) {
			 System.out.println("*****************************************");
			 System.out.println("BEGIN of the ChatMessageBroadcast TREATMENT");
			 ChatMessageBroadcast c = (ChatMessageBroadcast) o;
			 System.out.println("ID: " + c.getId());
			 System.out.println("Txt " + c.getTxt());
			// BROADCAST communication model setted
			notifyToAllOtherPlayers(o, senderASID);
			 System.out.println("END of the ping PLAYER TREATMENT");
			 System.out.println("*****************************************"); 
		}
		else if (o instanceof ChatMessageMulticast) {
			 System.out.println("*****************************************");
			 System.out.println("BEGIN of the ChatMessageMulticast TREATMENT");
			 ChatMessageMulticast c = (ChatMessageMulticast) o;
			 System.out.println("ID: " + c.getId());
			 System.out.println("Txt " + c.getTxt());
			// MULTICAST communication model setted
			notifyToRecipients(o, ((ChatMessageMulticast) o).getRecipients());
			 System.out.println("END of the ChatMessageMulticast TREATMENT");
			 System.out.println("*****************************************"); 
		}
		else if (o instanceof ChatMessageUnicast) {
			 System.out.println("*****************************************");
			 System.out.println("BEGIN of the ChatMessageUnicast TREATMENT");
			 ChatMessageUnicast c = (ChatMessageUnicast) o;
			 System.out.println("ID: " + c.getId());
			 System.out.println("Txt " + c.getTxt());
			 System.out.println("Recipient: " + c.getRecipient());
			// UNICAST communication model setted
			sendMessageToPlayer(o, ((ChatMessageUnicast) o).getRecipient());
			System.out.println("END of the ChatMessageUnicast TREATMENT");
			System.out.println("*****************************************"); 
		}
		//#ifdef REAL_WORLD_MAP
		else if (o instanceof LocationEvent) {
			System.out.println("*****************************************");
			System.out.println("BEGIN of the LocationEvent TREATMENT");
			LocationEvent l = (LocationEvent) o;
			p = (Player) playersTable.get(""+l.getId());
			
			p.setLatitude(l.getLatitude());
			p.setLongitude(l.getLongitude());
			
			mapPositionProjectionToOtherPlayers(l);
			
			System.out.println("END of the LocationEvent TREATMENT");
			System.out.println("*****************************************"); 
		}
		else if (o instanceof MapRequest) {
			System.out.println("*****************************************");
			System.out.println("BEGIN of the MapRequest TREATMENT");
			MapRequest mapRequest = (MapRequest) o;
			p = (Player) playersTable.get(""+mapRequest.getId());
			
			p.setLatitude(mapRequest.getLatitude());
			p.setLongitude(mapRequest.getLongitude());
			p.setMapWidth(mapRequest.getWidth());
			p.setMapHeight(mapRequest.getHeight());
			p.setMapCenterLatitude(mapRequest.getLatitude());
			p.setMapCenterLongitude(mapRequest.getLongitude());
			p.setZoom(mapRequest.getZoom());
			
			mapPositionProjections(mapRequest.getId(),mapRequest.getZoom());
			
			System.out.println("END of the MapRequest TREATMENT");
			System.out.println("*****************************************"); 
		}
		//#endif REAL_WORLD_MAP
	}


	//#ifdef REAL_WORLD_MAP
	private Object projectionMonitor = new Object();
	
	private void mapPositionProjections(int id, int zoom) {
		synchronized(projectionMonitor){
			Player requester = (Player) playersTable.get(""+id);
			Player p;
			MapResponse mapResponse = new MapResponse();
			PlayerCoordinates[] allCoords = new PlayerCoordinates[playersTable.size()-1];
			
			// init location library centered on requester position
			projection.initLL2XY(requester.getLatitude(), requester.getLongitude(), requester.getMapWidth(), requester.getMapHeight(), zoom);
			System.out.println("projection initiated for player n�"+(short) id+": center "+requester.getLatitude()+"/"+requester.getLongitude()+" width/height "+requester.getMapWidth()+"/"+requester.getMapHeight()+" zoom:"+zoom);
			
			int index = 0;
			Enumeration enume = playersTable.keys();
			while (enume.hasMoreElements()){
				short pid = Short.parseShort((String) enume.nextElement());
				if (id != pid){
					// compute position projection for other players
					p = (Player) playersTable.get(""+pid);
					
					// no location for player continue
					if (!p.locationReceived) continue;
					
					// adjust openmap projection for yahoomaps
					double adjustedLat = projection.errorAdjustement(p.getLatitude());  
					
					// LLXY projection
					Point proj = projection.LL2XYProjection(adjustedLat, p.getLongitude());
					
					PlayerCoordinates coords = new PlayerCoordinates();
					coords.setId((short) pid);
					coords.setX(proj.x);
					coords.setY(proj.y);
					
					System.out.println("projected coordinates for player n�"+(short) pid+": "+proj.x+"/"+proj.y);
					
					allCoords[index] = coords;
					index++;
				}
			}
			mapResponse.setPlayerCoordinates(allCoords);
			sendMessageToPlayer(mapResponse,id);
		}
	}
	
	private void mapPositionProjectionToOtherPlayers(LocationEvent location) {
		synchronized(projectionMonitor){
			Player requester = (Player) playersTable.get(""+location.getId());
			Player p;
			
			Enumeration enume = playersTable.keys();
			while (enume.hasMoreElements()){
				short pid = Short.parseShort((String) enume.nextElement());
				if (location.getId() != pid){
					// compute position projection for other players
					p = (Player) playersTable.get(""+pid);
					
					// no location for player continue
					if (!p.mapRequestReceived) continue;
					
					// init location library centered on requester position
					projection.initLL2XY(p.getMapCenterLatitude(), p.getMapCenterLongitude(), p.getMapWidth(), p.getMapHeight(), p.getZoom());
					
					// adjust openmap projection for yahoomaps
					double adjustedLat = projection.errorAdjustement(location.getLatitude());  
					
					// LLXY projection
					Point proj = projection.LL2XYProjection(adjustedLat, location.getLongitude());
					
					System.out.println("projected coordinates of player "+location.getId()+" map width/height "+p.getMapWidth()+"/"+p.getMapWidth()+" for player n�"+(short) pid+": "+proj.x+"/"+proj.y);
					
					location.setProjectedX(proj.x);
					location.setProjectedY(proj.y);
					
					sendMessageToPlayer(location,pid);
				}
			}
		}
	}
	//#endif REAL_WORLD_MAP
	
	//#ifdef PING
	//# private void sendEachOtherPlayersPings(int id, long time) {
	//# 
		//# Enumeration enume=htPlayerPing.keys();
		//# while (enume.hasMoreElements()){
			//# short pid = Short.parseShort((String) enume.nextElement());
			//# PingPlayer ppl = (PingPlayer) htPlayerPing.get(""+pid);
			//# int desync = (int)(time -ppl.getTimestamp());
			//# ppl.setDesynchronizationWithSender(desync);
		//# }
		//# sendMessagesToPlayer (htPlayerPing,id);
		//# if (!htPlayerPing.isEmpty()){
			//# short pid = Short.parseShort((String) enume.nextElement());
			//# PingPlayer ppl = (PingPlayer) htPlayerPing.get(""+pid);
			//# htPlayerPing.remove(ppl);
		//# }
	//# }
		//# 
	//# /**
	 //# * Add a PingPlayer object to vectors of each other players
	 //# * 
	 //# * @param ppl the PingPlayer object to store
	 //# */
	//# private void addPingMessageToEachOtherPlayers(PingPlayer ppl){
		//# Enumeration enume = htPlayerPing.keys();
		//# while (enume.hasMoreElements()){
			//# Hashtable playerPing = new Hashtable();
			//# short pid = Short.parseShort((String) enume.nextElement());
			//# while (enume.hasMoreElements()){
				//# if (pid != ppl.getId()){
					//# playerPing.put(""+pid,ppl);
				//# }
			//# }
			//# htPlayerPing.put(""+pid,playerPing);
		//# }
		//# 
	//# }
	//#endif PING

	/**
	 * Sending to the player (aSID) the object ErrorMessage with the type type
	 * 
	 * @param type 
	 * @param aSID
	 */
	
	private void sendErrorMessage(byte type, int aSID) {
		System.out.println("Error message sended, type: " + type);
	}
	
	
	/**
	 * Send to the player (idJoiner) a newPlayer message for each player in game,
	 * whose all informations were recovered.
	 * Only automatic mode
	 * 
	 * @param idJoiner
	 */
	
	private void sendInGamePlayersInfos(int idJoiner) {
		// synchronized (monitorActors){
		System.out.println("sendInGamePlayersInfos called");
		Hashtable h = new Hashtable();
		int nbObj = 0;
		Player p = null;
		showPlayers();
		
		// Build a newPlayer message to all players already in game and add in the hashtable.
		Enumeration enume = playersTable.keys();
		while (enume.hasMoreElements()){
			short pid = Short.parseShort((String) enume.nextElement());
			if (idJoiner != pid){
				p = (Player) playersTable.get(""+pid);
				System.out.println("playerAimed(" + idJoiner + ") playerInfos(" + pid
						+ ")");
				
				// new player message to send to idJoiner
				NewPlayer np = new NewPlayer();
				np.setId((short) p.getId());
				np.setX(p.getX());
				np.setY(p.getY());
				np.setUsername(p.getUsername());
				h.put(""+nbObj,np);
				
				nbObj++;
				System.out.println("NEW PLAYER message added to"+ idJoiner
						+ ": player n�"+ p.getId());
			}
		}
		
		// Build a dataEvent if some newPlayer messages have to be sent.
		if (!h.isEmpty()) {
			DataEvent de = new DataEventImpl(0, h);
			System.out.println("SendDataTo player>" + idJoiner + " de>" + de);
			this.sendDataTo(idJoiner, de);
		}
		// }
	}

	/**
	 * Sending a Message Object to all players
	 * @param o
	 *  
	 */
	private void notifyToAllPlayers(Object o) {
		System.out.println("notifyAllPlayers called");
		Hashtable h = new Hashtable();
		h.put("0", o);

		DataEvent de = new DataEventImpl(0, h);
		showPlayers();

		Enumeration enume= playersTable.keys();
		while (enume.hasMoreElements()){
			short pid = Short.parseShort((String) enume.nextElement());
			this.sendDataTo(pid,de);
		}
	}
		

	/**
	 * Sending a Message object to each player but not to the sender.
	 * 
	 * @param o, idSender
	 */
	private void notifyToAllOtherPlayers(Object o, int idSender) {
		System.out.println("notifyAllOtherPlayers called");
		Hashtable h = new Hashtable();
		h.put("0", o);

		DataEvent de = new DataEventImpl(0, h);
		showPlayers();
		
		Enumeration enume= playersTable.keys();
		while (enume.hasMoreElements()){
			short pid = Short.parseShort((String) enume.nextElement());
			if ((idSender != pid)){
				this.sendDataTo(pid,de);
			}
		}
	}
	
	
	private void notifyToRecipients(Object o,short[] recipients) {
		System.out.println("notifyToRecipients called");
		Hashtable h = new Hashtable();
		h.put("0",o);
		DataEvent de = new DataEventImpl(0,h);
		int i = 0;
		while (i<recipients.length) {
			System.out.println("Recipient: " + recipients[i]);
			if (playersTable.get(("" + recipients[i])) != null) {
				System.out.println("send to recipient => " + recipients[i]);
				sendDataTo(recipients[i],de);
			}
			i++;
		}
	}


	/**
	 * Sending a Message object to a specific player.
	 * 
	 * @param o, playerASID
	 */
	private void sendMessageToPlayer(Object o, int playerASID) {
		if (o instanceof Hashtable){
			Hashtable ht = (Hashtable) o;
			System.out.println("sendMessageToPlayer called");
			Hashtable h = new Hashtable();
			
			int count = 0;
			Enumeration enume = ht.keys();
			while (enume.hasMoreElements()){
				short pid = Short.parseShort((String) enume.nextElement());
				h.put(""+count,ht.get(""+pid));
				count++;
			}
			/*for(Iterator iter = v.iterator(); iter.hasNext();){
				h.put(""+count, iter.next());
				count++;
			}*/
			DataEvent de = new DataEventImpl(0, h);
			this.sendDataTo(playerASID, de);
		}
		else{
			System.out.println("sendMessageToPlayer called");
			Hashtable h = new Hashtable();
			h.put("0", o);
	
			DataEvent de = new DataEventImpl(0, h);
			this.sendDataTo(playerASID, de);
		}
	}

	/**
	 * Show players state
	 *
	 */
	private void showPlayers() {
		Enumeration enume= playersTable.keys();
		while (enume.hasMoreElements()){
			short pid = Short.parseShort((String)enume.nextElement());
			System.out.println("playersTable key{"+ pid+"} = "+ playersTable.get(""+pid));
		}
	}
}
